/*
 *  TODO:
 *
 *  1) add timeout to connect and recv
 *  2) socket file locking
 */

//#define __DEBUG__

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <getopt.h>
#include <lib/util.h>
#include <router/ipqext/ipqextctl.h>

const char *program_version = "1.0";
const char *program_name = "ipqext";

static struct option opts[] = {
  {"add",     required_argument, 0, 'A'},
  {"delete",  required_argument, 0, 'D'},
  {"move",    required_argument, 0, 'M'},
  {"start",   required_argument, 0, 'S'},
  {"stop",    required_argument, 0, 'X'},
  {"group",   required_argument, 0, 'G'},
  {"reset",   optional_argument, 0, 'Z'},
  {"info",    required_argument, 0, 'i'},
  {"flush",   no_argument,       0, 'F'},
  {"list",    no_argument,       0, 'L'},
  {"mark",    required_argument, 0, 'm'},
  {"help",    optional_argument, 0, 'h'},
  {0, 0, 0, 0}
};

static const char cmdflags[] = 
  { 'A', 'D', 'M', 'S', 'X', 'G', 'Z', 'i', 'F', 'L', 'm', 'h'};

#define ERR_INVALID_CMD 1

unsigned int command;

static char cmd2char(int cmd)
{
  int i;
  for (i=0; (i < (sizeof(int)<<3)) && ((cmd & 1) == 0); cmd >>=1, i++);
  return cmdflags[i];
}

static void exit_tryhelp()
{
	fprintf(stderr, "Try `%s -h' or '%s --help' for more information.\n",
          program_name, program_name);
}

static void exit_error(int status, char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	fprintf(stderr, "%s v%s: ", program_name, program_version);
  if (msg != NULL)
    vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");

  if (status == ERR_INVALID_CMD) {
    exit_tryhelp();
  }

	exit(status);
}

static void exit_printhelp(void)
{
	printf("%s v%s\n\n"
         "Usage: %s -[A] extension [parameters]\n"
         "       %s -[DSXih] extension\n"
         "       %s -M extension index\n"
         "       %s -G extension groupid\n"
         "       %s -Z [extension]\n"
         "       %s -F\n"
         "       %s -L\n"
         "       %s -m extension-list\n"
         "\n",
	       program_name, program_version,
         program_name, program_name, program_name,
         program_name, program_name, program_name,
         program_name, program_name);

	printf("Commands:\n"
         "Either long or short options are allowed.\n"
         "  --add     -A extension          Add an extension\n"
         "  --delete  -D extension          Delete an extension\n"
         "  --move    -M extension index    Move an extension to the specific index\n"
         "  --group   -G extension groupid  Set the group ID of an extension\n"
         "  --start   -S extension          Start an extension\n"
         "  --stop    -X extension          Stop an extension\n"
         "  --reset   -Z [extension]        Reset internal data of an extension or\n"
         "                                  all extensions\n"
         "  --info    -i extension          Show the information of an extension\n"
         "  --flush   -F                    Delete all extensions\n"
         "  --list    -L                    List all extensions\n"
         "  --mark    -m extension-list     Get the netfilter mark value for a list of\n"
         "                                  extensions (extension-list is a list of\n"
         "                                  comma-separated extension names)\n"
         "  --help    -h [extension]        Print this help or the help information of\n"
         "                                  an extension\n"
         "\n");
  exit(0);
}

static void add_command(int *cmd, const int newcmd)
{
	if (*cmd != 0) {
    exit_error(1, "Can't use -%c with -%c\n",
               cmd2char(newcmd), cmd2char(*cmd));
  }
	*cmd |= newcmd;
}


static void parse_options(ipqextctl_request_t *request, int argc, char **argv)
{
  int optidx = 0;
  int c;
  char *extname = NULL, *extlist = NULL;

  memset(request, 0, sizeof(ipqextctl_request_t));

  opterr = 0;
	while ((c = getopt_long(argc, argv, "A:D:M:G:S:X:Z::i:FLm:h::", opts, &optidx)) != -1) {
		switch (c) {
    case 0:
      break;

    case 'A':
      add_command(&(request->command), IPQEXTCTL_CMD_ADD);
      extname = optarg;
      break;

    case 'D':
      add_command(&(request->command), IPQEXTCTL_CMD_DEL);
      extname = optarg;
      break;

    case 'M':
      add_command(&(request->command), IPQEXTCTL_CMD_MOVE);
      extname = optarg;
      break;

    case 'G':
      add_command(&(request->command), IPQEXTCTL_CMD_GROUP);
      extname = optarg;
      break;

    case 'S':
      add_command(&(request->command), IPQEXTCTL_CMD_START);
      extname = optarg;
      break;

    case 'X':
      add_command(&(request->command), IPQEXTCTL_CMD_STOP);
      extname = optarg;
      break;

    case 'Z':
      add_command(&(request->command), IPQEXTCTL_CMD_RESET);
			if (optarg)
        extname = optarg;
			else if (optind < argc && argv[optind][0] != '-')
				extname = argv[optind++];
      else
        extname = NULL;
      break;

    case 'i':
      add_command(&(request->command), IPQEXTCTL_CMD_INFO);
      extname = optarg;
      break;

    case 'F':
      add_command(&(request->command), IPQEXTCTL_CMD_FLUSH);
      extname = NULL;
      break;

    case 'L':
      add_command(&(request->command), IPQEXTCTL_CMD_LIST);
      extname = NULL;
      break;

    case 'm':
      add_command(&(request->command), IPQEXTCTL_CMD_MARK);
      extname = NULL;
      extlist = optarg;
      break;

    case 'h':
      add_command(&(request->command), IPQEXTCTL_CMD_HELP);
			if (optarg)
        extname = optarg;
			else if (optind < argc && argv[optind][0] != '-')
				extname = argv[optind++];
      else
        exit_printhelp();      
      break;

    case '?':
      if (isprint(optopt))
        exit_error(ERR_INVALID_CMD,
                   "Unknown option or missing argument for switch `-%c'.",
                   optopt);
      else
        exit_error(ERR_INVALID_CMD,
                   "Unknown option character `\\x%x'.", optopt);
      break;

    default:
      exit_error(ERR_INVALID_CMD, "Invalid command.");
    }
  }

  // no command
  if (request->command == 0)
    exit_printhelp();      

  // no extension name
  if (extname != NULL) {
    if (extname[0] == '-')
      exit_error(ERR_INVALID_CMD, "Invalid extension name.");    
    strncpy(request->extname, extname, IPQEXT_NAME_LEN);
  }

  // parse extra options
  switch (request->command) {
  case IPQEXTCTL_CMD_ADD:
    if (optind >= argc-1) {
      if (optind == argc-1)
        memcpy(request->param, argv[optind],
               min(IPQEXT_PARAM_LEN, strlen(argv[optind])));
    } else {
      exit_error(ERR_INVALID_CMD, "Invalid argument.");
    }
    break;
    
  case IPQEXTCTL_CMD_MOVE:
  case IPQEXTCTL_CMD_GROUP:
    if (optind == argc-1) {
      memcpy(request->param, argv[optind],
             min(IPQEXT_PARAM_LEN, strlen(argv[optind])));
    } else {
      exit_error(ERR_INVALID_CMD, "Invalid argument.");
    }
    break;

  case IPQEXTCTL_CMD_MARK:
    if (optind == argc) {
      memcpy(request->param, extlist,
             min(IPQEXT_PARAM_LEN, strlen(extlist)));      
    } else {
      exit_error(ERR_INVALID_CMD, "Invalid argument.");    
    }

  default:
    if (optind != argc)
      exit_error(ERR_INVALID_CMD, "Invalid argument.");
  }
}

int ctlsd;
struct sockaddr_un saddr;

static int ipqext_init()
{
  // initialize socket
  DB("initializing ipqextctl ...\n");
  ctlsd = IPQEXTCTL_INIT_TCP_SOCKET(IPQEXT_SOCKET_PATHNAME);
  if (ctlsd == -1) {
    return -1;
  }

  ipqextctl_set_sockaddr(&saddr, IPQEXTD_SOCKET_PATHNAME);

  // connect to ipqextd
  DB("connecting to ipqextd ...\n");
  if (connect(ctlsd, (struct sockaddr *) &saddr,
              sizeof(struct sockaddr_un)) == -1) {
    ERR_RET(-1);
  }

  return 0;
}

static void ipqext_fini()
{
  // close connection
  DB("closing connection ...\n");
  if (shutdown(ctlsd, SHUT_WR) == -1) {
    ERR();
  }
  close(ctlsd);
}

static int ipqext_send(ipqextctl_request_t *request)
{
  ipqextctl_response_t response;
  int recvlen, finished;

  // send command
  DB("sending command to ipqextd ...\n");
  if (send(ctlsd, request, sizeof(ipqextctl_request_t), 0) == -1) {
    ERR_RET(-1);
  }

  // wait for response
  finished = 0;
  while (!finished) {
    recvlen = recv(ctlsd, &response, sizeof(ipqextctl_response_t), 0);
    if (recvlen == sizeof(ipqextctl_response_t)) {
      switch (response.status) {
      case IPQEXTCTL_STAT_OK:
        if (strlen(response.param) > 0) {
          printf("%s\n", response.param);
        }
        finished = 1;
        break;
        
      case IPQEXTCTL_STAT_CONT:
        printf("%s\n", response.param);
        break;
        
      case IPQEXTCTL_STAT_FAIL:
      case IPQEXTCTL_STAT_UNKNOWN:
        printf("ERROR: %s\n", response.param);
        finished = 1;
        break;
      }
    } else {
      DB("received length %d\n", recvlen);
      ERR_RET(-1);
    }
  }

  return 0;
}

int main(int argc, char **argv)
{
  ipqextctl_request_t request;

  parse_options(&request, argc, argv);
  
  if (ipqext_init() == -1)
    exit(1);

  ipqext_send(&request);
  ipqext_fini();

  return 0;
}
